// $Id: AlarmClock.cc,v 1.10.2.1 1998/09/22 07:43:19 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AlarmClock.h"
#include "Log.h"

AlarmClock TheAlarmClock;

// threshold to warn about late alarms
static Time TheMaxReportedDelay(-2,0); // late => negative time diff

SafePtrDefn(SleeperPtr, Sleeper)

/* Alarm */

int Alarm::theLastId = 0;

void Alarm::ring() const {
	Notepad << "alarm[" << theId << "]: " << theSleeper << "->" << "wakeUp" << '(' << theId << ')' << ende;
	theSleeper->wakeUp(*this);
}

void Alarm::set(Time aTime, Sleeper *aSleeper) {
	Assert(aSleeper);
	theTime = aTime;
	theSleeper.assign(aSleeper->useRec(), aSleeper);
}

/* Sleeper */

void Sleeper::sleepTill(Time time) {
	TheAlarmClock.setAlarm(Alarm(time, this));
}

void Sleeper::sleepFor(Time delay) {
	sleepTill(TheClock + delay);
}

/* AlarmQueue */

// handy to skip one item, detroying it
inline static
void advance(AlarmQueueItem *cur, AlarmQueueItem *&next) {
	next = cur->theNext;
	delete cur;
}

AlarmQueue::~AlarmQueue() {
	while (theHead)
		shift();
}

void AlarmQueue::enqueue(const Alarm &a) {
	AlarmQueueItem **cur = &theHead;
	while (*cur && (*cur)->theAlarm.time() <= a.time())
		cur = &(*cur)->theNext;
	*cur = new AlarmQueueItem(a, *cur);
}

// find first valid alarm
Alarm *AlarmQueue::first() {
	while (theHead && !theHead->theAlarm.isSet())
		advance(theHead, theHead);
	return theHead ? &theHead->theAlarm : 0;
}

// dequeues and destroys first alarm
void AlarmQueue::shift() {
	if (theHead) {
		AlarmQueueItem *oldHead = theHead;
		theHead = oldHead->theNext;
		delete oldHead;
	}
}

/* AlarmClock */

void AlarmClock::setAlarm(const Alarm &alarm) { 
	if (alarm.isSet()) {
		if (alarm.time() <= TheClock) { // already ready
			alarm.ring();
		} else {
			theQueue.enqueue(alarm);
			if (theRingTime <= 0 || alarm.time() < theRingTime) {
				noteNewRingTime(alarm.time());
			}
			Notepad << "alarm[" << alarm.id() << "] for " << alarm.sleeper() << " set at " << alarm.time() << " in " << (alarm.time() - TheClock) << ende;
		}
	}	
}

void AlarmClock::ring() {
	Alarm *alarm;
	while ((alarm = theQueue.first())) {
		noteNewRingTime(alarm->time());
		if (timeLeft() <= 0) {
			Notepad << "alarm[" << alarm->id() << "] for " << alarm->sleeper() << " rings with a delay of " << timeLeft() << ende;
			// warn about too late/early alarms
			if (timeLeft() < TheMaxReportedDelay) {
				SysLog << "alarm[" << alarm->id() << "] is late for at least " << timeLeft() << ende;
				TheMaxReportedDelay = timeLeft();
			}
			alarm->ring();
			theQueue.shift();
		} else {
			break;
		}
	}
}

void AlarmClock::update(Time curTime) {
	Clock::update(curTime);
	theTimeLeft = theRingTime - curTime;
	if (theTimeLeft <= 0)
		ring();
}

void AlarmClock::noteNewRingTime(Time time) {
	theTimeLeft = time - theCurTime;
	theRingTime = time;
}
