// $Id: IOBuffer.cc,v 1.6.2.1 1998/09/22 07:43:28 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "IOBuffer.h"

#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "Bug.h"

/*
              /offset                    /capacity
    |---------*****************---------|
    <----     size        ---->
*/


IOBuffer::IOBuffer(): theBuf(0),
	theCapacity(0), theSize(0), theOffset(0),
	theInOffset(0), theOutOffset(0),
	theState(0),
	theError(0), theFmtBug(false), theEOD(false) {
}

IOBuffer::~IOBuffer() {
	delete[] theBuf;
}

void IOBuffer::reset() {
	theSize = 0;
	theOffset = theInOffset = theOutOffset = 0;
	theError = 0;
	theFmtBug = theEOD = false;
	theState = 0;
	if (theBuf)
		*theBuf = '\0';
}

void IOBuffer::append(const char *data, Size size) {
	stretch(size);
	assert(theSize + size <= theCapacity);
	memcpy(theBuf + theSize, data, size);
	appended(size);
}

// makes bufSize() >= sizeToFit
void IOBuffer::stretch(Size sizeToFit) {
	if (theSize + sizeToFit <= theCapacity)
		return; /* fits in the tail */
	else
	if (contSize() + sizeToFit > theCapacity)
		grow(contSize() + sizeToFit); /* will not fit anyhow */
	else
		pack(); /* fits if packed */
}

void IOBuffer::pack() {
	if (theOffset > 0) {
		if (contSize())
			memmove(theBuf, theBuf + theOffset, contSize());
		theSize -= theOffset;
		theOffset = 0;
		memset(theBuf+theSize, 0, theCapacity-theSize);
	}
}

/* convenience operators */

// borrowed from libg++/libio/iostream.cc with simplifications
IOBuffer &IOBuffer::operator <<(int n) {
#	define WRITE_BUF_SIZE (10 + sizeof(int) * 3)
    char buf[WRITE_BUF_SIZE];
    register char *buf_ptr = buf+WRITE_BUF_SIZE; // end of buf.
	do {
		*--buf_ptr = (n % 10) + '0';
		n /= 10;
	} while (n != 0);
	append(buf_ptr, buf+WRITE_BUF_SIZE - buf_ptr);
	return *this;
}

IOBuffer &IOBuffer::operator >>(char &c) {
	if (errFmt() || contSize() <= 0) {
		errFmt(true);
	} else {
		c = *content();
		consumed(1);
	}
	return *this;
}

IOBuffer &IOBuffer::operator >>(char c) {
	if (errFmt() || contSize() <= 0)
		errFmt(true);
	else
	if (*content() != c)
		errFmt(true);
	else
		consumed(1);
	return *this;
}

// extracts characters into s till the end of buffer or a white space
IOBuffer &IOBuffer::operator >>(char *s) {
	if (errFmt() || contSize() <= 0) {
		errFmt(true);
		s[0] = '\0';
	} else {
		const char *buf = content();
		int buf_size = contSize();
		const char *black = buf;
		int len = 0;
		// read till end of buffer or white space
		while (*black && !isspace(*black) && len < buf_size) {
			black++;
			len++;
		}
		if (len) {
			memcpy(s, buf, len);
			consumed(len);
		}
		s[len] = '\0';
	}
	return *this;
}

// tests if a given string is there (case-insensitive)
IOBuffer &IOBuffer::operator >>(const char *s) {
	if (errFmt())
		return *this;
	int len = strlen(s);
	if (contSize() < len || strncasecmp(content(), s, len) != 0) {
		errFmt(true);
		return *this;
	}
	// match found, extend pointer
	consumed(len);
	return *this;
}

IOBuffer &IOBuffer::operator >>(int &n) {
	if (errFmt())
		return *this;
	const char *buf = content();
	char *end_ptr;
	n = (int)strtol(buf, &end_ptr, 10);
	if (end_ptr > buf)
		consumed(end_ptr - buf);
	else
		errFmt(true);
	return *this;
}

void IOBuffer::skipWhiteSpace() {
	if (!errFmt()) {
		int len = contSize();
		register const char *buf = content();
		while (*buf && isspace(*buf) && len-- > 0) buf++;
		consumedUpTo(buf);
	}
}

/* low-level operations */

// note: theIn/OutOffset's are not reset!
void IOBuffer::clean() {
	theOffset = theSize = 0; 
	if (theBuf)
		memset(theBuf, 0, theCapacity);
}

// this function ensures that theBuf is 0-terminated
// the price is up to one "chunk" of not-really-needed memory
void IOBuffer::grow(Size neededCapacity) {
	static const Size chunkSize = Size::KB(4);

	neededCapacity += chunkSize; /* not chunkSize-1, because of 0-termination */
	neededCapacity /= chunkSize;
	neededCapacity *= chunkSize;

	char *buf = new char [neededCapacity];
	memset(buf, 0, neededCapacity);

	if (contSize())
		memcpy(buf, theBuf+theOffset, contSize());
	if (theBuf)
		delete[] theBuf;

	theBuf = buf;
	theCapacity = neededCapacity - 1; /* because of 0-termination */
	theSize -= theOffset;
	theOffset = 0;
}
