// $Id: Shooter.cc,v 1.3.2.1 1998/09/22 07:43:42 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Shooter.h"
#include "Gadgets.h"

#include <iostream.h>
#include <iomanip.h>

ShotRegistry *TheShotRegistry = 0;

static const char *TheShotRegistryMagic = "ShotRegistry";

/* Shooter */

Shooter::Shooter(Name aName, Size aSize, char *aBody):
	theName(aName), theSize(aSize), theBody(aBody) {

	assert(TheShotRegistry);
	TheShotRegistry->add(this);
}

void Shooter::shoot(char *buf) {
	assert(theBody);
	memcpy(buf, theBody, theSize);
	reset();
}

void Shooter::restore(const char *buf) {
	assert(theBody);
	memcpy(theBody, buf, theSize);
}

void Shooter::reset() {
	assert(theBody);
	memset(theBody, 0, theSize);
}


/* ShotRegistry */

ShotRegistry::ShotRegistry(): theShotSize(sizeof(int)), theShotCount(0) {
	assert(sizeof(int) == 4);    // hton32 and ntoh32 depend on it
	assert(sizeof(Time) == 4+4); // Time manipulations depends on it
}

ShotRegistry::~ShotRegistry() {
	for (int i = 0; i < theShots.count(); ++i)
		delete[] theShots[i];
}

void ShotRegistry::add(Shooter *shooter) {
	assert(!theShotCount);
	theShooters.append(shooter);
	theShotSize += shooter->size();
}

// allocates count more buffers for shots
void ShotRegistry::preAlloc(int count) {
	while (count-- > 0)
		addShotBuf();
}

// allocates one more buffer for shots
void ShotRegistry::addShotBuf() {
	char *buf = new char[theShotSize];
	assert(buf);
	theShots.append(buf);
}

void ShotRegistry::snapShot(Time curTime) {
	assert(theFirstShot > 0);

	if (theShotCount >= theShots.count())
		addShotBuf();
	assert(theShotCount <= theShots.count());

	char *buf = theShots[theShotCount];
	Size off = 0;

	const int delay = (curTime - theLastShot).usec();
	memcpy(buf, &delay, sizeof(delay));
	off += sizeof(delay);

	for (int i = 0; i < theShooters.count(); ++i) {
		Shooter *s = theShooters[i];
		s->shoot(buf + off);
		off += s->size();
	}
	assert(off == theShotSize);
	theLastShot = curTime;
	theShotCount++;
}

void ShotRegistry::save(ostream &os) const {

	saveHeader(os);

	for (int b = 0; b < theShotCount; ++b) {
		char *shot = theShots[b];
		Size off = 0;

		// convert to network byteorder

		// time delta
		hton32(shot);
		off += 4;

		for (int i = 0; i < theShooters.count(); ++i) {
			const Shooter *s = theShooters[i];
			s->hton(shot + off);
			off += s->size();
		}

		// save
		os.write(shot, theShotSize);
	}
}

void ShotRegistry::print(istream &is, ostream &os) {

	if (loadHeader(is) < 0)
		return; // header format failure

	printHeader(os);

	// print table header
	os << "# dTm";
	for (int i = 0; i < theShooters.count(); ++i)
		os << " \t" << theShooters[i]->name();
	os << endl;

	addShotBuf();
	char *shot = theShots[0];
	for (int b = 0; b < theShotCount; ++b) {
		// load shot
		is.read(shot, theShotSize);

		// convert to host byteorder
		Size off = 0;

		// time delta
		ntoh32(shot);
		off += 4;

		for (int i = 0; i < theShooters.count(); ++i) {
			const Shooter *s = theShooters[i];
			s->ntoh(shot + off);
			off += s->size();
		}

		// print
		off = 0;

		// print delta
		int delta;
		memcpy(&delta, shot, sizeof(delta));
		os << setw(6) << delta;
		off += 4;

		for (int i = 0; i < theShooters.count(); ++i) {
			Shooter *s = theShooters[i];
			s->restore(shot + off);
			off += s->size();
			s->print(os << " \t");
			
		}
		os << endl;
	}
}


void ShotRegistry::saveHeader(ostream &os) const {
	// magic string
	os << TheShotRegistryMagic << endl;

	// save names
	os << theShooters.count() << endl;
	for (int i = 0; i < theShooters.count(); ++i) {
		os << theShooters[i]->name() << endl;
	}

	// other info
	os
		<< theShotSize << endl
		<< theFirstShot << endl
		<< theLastShot << endl
		<< theShotCount << endl
		<< ends;
}

int ShotRegistry::loadHeader(istream &is) {
	assert(!theShotCount);

	// magic string
	if (!GetConstLine(is, TheShotRegistryMagic))
		return -1;

	int count;

	// load names
	is >> count >> ws;
	if (count != theShooters.count())
		return -1; // count must be the same
	for (int i = 0; i < theShooters.count(); ++i) {
		if (!GetConstLine(is, theShooters[i]->name()))
			return -1; // order must be the same
	}

	// other info
	char c = '?';
	is
		>> theShotSize >> ws
		>> theFirstShot >> ws
		>> theLastShot >> ws
		>> theShotCount >> ws
		>> c;
	if (c != '\0')
		return -1; /* must end with ends */

	// success
	return theShotCount;
}

void ShotRegistry::printHeader(ostream &os) const {
	os << '#'
		<< " shooters: " << theShooters.count()
		<< " shot size: " << theShotSize << " bytes,"
		<< " count: " << theShotCount
		<< " volume: " << diskSize() << " bytes"
		<< endl;
	os << '#'
		<< " start: " << theFirstShot
		<< " end: " << theLastShot
		<< " (" << (theLastShot - theFirstShot) << ")"
		<< endl;
}



/* ShotRegistryLibCounter; internal stuff, ignore */

int ShotRegistryLibCounter::theUseCount = 0;

ShotRegistryLibCounter::ShotRegistryLibCounter() {
	if (!theUseCount++)
		TheShotRegistry = new ShotRegistry();
}

ShotRegistryLibCounter::~ShotRegistryLibCounter() {
	if (!--theUseCount) {
		delete TheShotRegistry;
		TheShotRegistry = 0;
	}
}
