#ifndef _ALARM_CLOCK_H_
#define _ALARM_CLOCK_H_ "$Id: AlarmClock.h,v 1.11.2.1 1998/09/22 07:43:55 rousskov Exp $"

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Array.h"
#include "Clock.h"
#include "SafePtr.h"
#include "Identity.h"

class Sleeper;
SafePtrDecl(SleeperPtr, Sleeper)

// one time alarm
class Alarm {
	public:
		Alarm(): theId(++theLastId) {}
		Alarm(Time aTime, Sleeper *aSleeper): theId(++theLastId) { set(aTime, aSleeper); }

		bool isSet() const { return theSleeper; }
		Time time() const { return theTime; }
		int id() const { return theId; }

		const Sleeper *sleeper() const { return theSleeper; }

		void set(Time aTime, Sleeper *aSleeper);
		void ring() const;
		void clear() { theSleeper = 0; }

	protected:
		Time theTime;
		SleeperPtr theSleeper;
		int theId;

	private:
		static int theLastId;
};

inline
ostream &operator <<(ostream &os, const Alarm &a) { return os << a.id(); }

// somebody who can wait for an alarm
// alarm count is used to check if pending alarms must be cleared 
//	before object destruction
class Sleeper: public virtual Identity, public virtual SelfAcc {
	public:
		Sleeper() {}

		virtual void wakeUp(const Alarm &alarm) = 0;

	protected:
		virtual void sleepTill(Time time);
		void sleepFor(Time delay);
};

class AlarmQueue;

// an atomic member of alarm queue
class AlarmQueueItem {
	public:
		AlarmQueueItem(const Alarm &anAlarm, AlarmQueueItem *aNext): theAlarm(anAlarm), theNext(aNext) {}

	public:
		Alarm theAlarm;
		AlarmQueueItem *theNext;
};

// a sequence of ordered AlarmQueueItems
class AlarmQueue {
	public:
		AlarmQueue(): theHead(0) {}
		~AlarmQueue();

		bool empty() const { return !theHead; }

		void enqueue(const Alarm &alarm);
		Alarm* first(); // the first valid alarm
		void shift();   // dequeues and destroys the first alarm

	protected:
		AlarmQueueItem *theHead;
};

// alarm clock that can handle unlimited number of alarms
class AlarmClock: public Clock {

	public:
		AlarmClock() {}
		~AlarmClock() {}

		bool isClear() const { return theQueue.empty(); }
		Time timeLeft() const { return theTimeLeft; }
		Time ringTime() const { return theRingTime; }

		virtual void update(Time curTime); // rings if needed
		void setAlarm(const Alarm &alarm);
		void ring(); // wakes up ready sleepers

	protected:
		void noteNewRingTime(Time time);

	protected:
		AlarmQueue theQueue;
		Time theTimeLeft;
		Time theRingTime;
};

extern AlarmClock TheAlarmClock;

#endif
