// $Id: tunnel.cc,v 1.4.2.1 1998/09/22 07:43:52 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include "IOBuffer.h"
#include "Stack.h"
#include "Socket.h"
#include "Gadgets.h"

#include <unistd.h>
#include <errno.h>
#include <string.h> /* for strerror */

#include <iostream.h>
#include <iomanip.h>


/*
 * Simple null-proxy
 * Tunnels traffic on a given towards given address
 */


// info for one direction of the IO stream (used by Side)
class DirRec {
	public:
		DirRec(): theBuf(0), isReady(false), isDone(true) {}

		bool done() const { return isDone; }
		bool ready() const { return isReady; }

		IOBuffer *buf() { return theBuf; }
		void buf(IOBuffer *aBuf) { theBuf = aBuf; isDone = isReady = false; }

		void beDone() { isDone = true; isReady = false; }
		void beReady(bool be = true) { isReady = be; }

	protected:
		IOBuffer *theBuf;
		bool isReady;     // fd is ready
		bool isDone;      // there should be no more IOs
};

// client or server side
class Side {
	public:
		Side(): theOtherSide(0), theFD(-1), theEodStr(0), theEodBuf(0), isOpen(false) { theEodBuf = new char[4096]; }
		~Side() { delete[] theEodBuf; }

		bool open() const { return isOpen; }
		bool alone() const { return !theOtherSide; }
		bool readReady() const { return theRead.ready(); }
		bool writeReady() const { return theWrite.ready(); }
		IOBuffer *readBuf() { return theRead.buf(); }
		IOBuffer *writeBuf() { return theWrite.buf(); }

		void fd(int aFD) { theFD = aFD; }
		void bufs(IOBuffer *aReadBuf, IOBuffer *aWriteBuf, const char *anEodStr = 0) { theRead.buf(aReadBuf); theWrite.buf(aWriteBuf); theEodStr = anEodStr; isOpen = true; }
		void otherSide(Side *anotherSide) { theOtherSide = anotherSide; }

		void read();
		void write();
		inline void close();

	protected:
		inline void doneAndCheck(DirRec &toBeDone, DirRec &toBeChecked);

	protected:
		DirRec theRead;
		DirRec theWrite;

		Side *theOtherSide;
		int theFD;

		const char *theEodStr; // string that indicates end-of-data on read
		char *theEodBuf;       // buffer to accumulate requests (XXX: kludge)

		bool isOpen;
};

// wrapper around fd_set
class FD_Set {
	public:
		FD_Set();

		int maxFD() const { return theMaxFD; }
		bool active() const { return theMaxFD >= 0; }

		bool isSetFD(int fd) const { return FD_ISSET(fd, &theSet); }
		bool isReadyFD(int fd) const { return FD_ISSET(fd, &theReadySet); }

		void setFD(int fd) { FD_SET(fd, &theSet); theCount++; if (fd > theMaxFD) theMaxFD = fd; }
		void clearFD(int fd) { if (isSetFD(fd)) { FD_CLR(fd, &theSet); FD_CLR(fd, &theReadySet); theCount--; if (fd == theMaxFD) while (--theMaxFD >= 0 && !isSetFD(theMaxFD)); } }

		fd_set *readySet() { if (!active()) return 0; memcpy(&theReadySet, &theSet, sizeof(theReadySet)); return &theReadySet; }

	protected:
		fd_set theSet;
		fd_set theReadySet;

		int theMaxFD;
		int theCount;
};

typedef enum IODir { ioRead = 0, ioWrite = 1 };

// wrapper around select(2)
class Select {
	typedef Stack<IOBuffer*> BufStack;
	public:
		Select();
		~Select();

		void limits(int aFDLimit, Size bufSize);

		int count() const { return theCount; }
		FD_Set &readSet() { return theSet[ioRead]; }
		FD_Set &writeSet() { return theSet[ioWrite]; }

		void addPair(int fd1, int fd2);

		void step(); // select and fill/empty buffers

		void addFD(int fd);
		void delFD(int fd);

	protected:
		Side *theSides;
		BufStack theBufs;

		FD_Set theSet[2];

		int theFDLimit;
		int theMaxFD;
		int theCount;
};

// globals
static const char *TheDestHost = 0;
static int ThePort = -1;
static int TheDestPort = -1;
static Select TheSelect;

/* Side */

void Side::read() {
	Assert(!theRead.done());
	IOBuffer &buf = *theRead.buf();
	buf.pack();
	const Size max_sz = buf.bufSize();
	if (max_sz > 0) {
		const ssize_t sz = ::read(theFD, buf.buf(), max_sz);
        buf.err(sz < 0 ? errno : 0);
        buf.eod(sz <= 0);
		//cerr << "fd: " << theFD << " read: " << sz << " from " << Socket(theFD).port() << (theRead.ready() ? " (auto)" : " (select)") << endl;
        Check(sz >= 0);
		if (sz > 0) {
			// XXX: kludge to stop reading early for HTTP requests
			//cerr << "content: " << buf.content() << endl;
			if (theEodStr) {
				Assert(buf.inOffset()+sz < 4096);
				strncpy(theEodBuf+buf.inOffset(), buf.buf(), sz);
				if (strstr(theEodBuf, theEodStr)) {
					buf.eod(true);
					doneAndCheck(theRead, theWrite);
				}
			}
            buf.appended(sz);
		} else { // sz == 0
			buf.eod(true);
			doneAndCheck(theRead, theWrite);
			if (theOtherSide && buf.exhausted())
				theOtherSide->doneAndCheck(theOtherSide->theWrite, theOtherSide->theRead);
		}
		if (theOtherSide && theOtherSide->writeReady())
			theOtherSide->write();
		if (theRead.ready()) { // were not called from select
			TheSelect.readSet().setFD(theFD); // re-register
			theRead.beReady(false);
		}
	} else 
	if (!theRead.ready()) { // were called from select
		theRead.beReady(true);
		TheSelect.readSet().clearFD(theFD);
	}
	//cerr << "fd: " << theFD << " after read: r: " << (theRead.done() ? "done" : theRead.ready() ? "ready" : "select") << endl;
	//cerr << "fd: " << theFD << " after read: w: " << (theWrite.done() ? "done" : theWrite.ready() ? "ready" : "select") << endl;
}

void Side::write() {
	Assert(!theWrite.done());
	IOBuffer &buf = *theWrite.buf();
	const Size max_sz = buf.contSize();
	if (max_sz > 0) {
		const ssize_t sz = ::write(theFD, buf.content(), max_sz);
		buf.err(sz < 0 ? errno : 0);
		//cerr << "fd: " << theFD << " wrote: " << sz << " to " << Socket(theFD).port() << (theWrite.ready() ? " (auto)" : " (select)") << endl;
		Check(sz >= 0);
		if (sz > 0) {
			buf.consumed(sz);
		}
		if (buf.exhausted())
			doneAndCheck(theWrite, theRead);
		if (theOtherSide && theOtherSide->readReady())
			theOtherSide->read();
		if (theWrite.ready()) { // were not called from select
			TheSelect.writeSet().setFD(theFD); // re-register
			theWrite.beReady(false);
		}
	} else 
	if (!theWrite.ready()) { // were called from select
		theWrite.beReady(true);
		TheSelect.writeSet().clearFD(theFD);
	}
	//cerr << "fd: " << theFD << " after write: w: " << (theWrite.done() ? "done" : theWrite.ready() ? "ready" : "select") << endl;
	//cerr << "fd: " << theFD << " after write: r: " << (theRead.done() ? "done" : theRead.ready() ? "ready" : "select") << endl;
}

inline
void Side::close() {
	TheSelect.delFD(theFD); // will close
	isOpen = false;
	theRead.buf(0);
	theWrite.buf(0);
	if (theOtherSide)
		theOtherSide->otherSide(0);
}

inline
void Side::doneAndCheck(DirRec &toBeDone, DirRec &toBeChecked) {
	if (!toBeDone.ready()) // may have registration
		(&toBeDone == &theRead) ? // XXX: ugly
			TheSelect.readSet().clearFD(theFD) :
			TheSelect.writeSet().clearFD(theFD);
	toBeDone.beDone();
	if (toBeChecked.done())
		close();
}

/* FD_Set */

FD_Set::FD_Set(): theMaxFD(-1), theCount(0) {
	FD_ZERO(&theSet);
	FD_ZERO(&theReadySet);
}

/* Select */

Select::Select(): theSides(0), theFDLimit(-1), theMaxFD(-1), theCount(0) {
}

Select::~Select() {
	delete[] theSides;
	while (theBufs.count()) delete theBufs.pop();
}

void Select::limits(int aFDLimit, Size bufSize) {
	Assert(theFDLimit < 0);
	theFDLimit = aFDLimit;
	theSides = new Side[theFDLimit];
	theBufs.stretch(theFDLimit);
	for (int i = 0; i < theFDLimit; ++i)
		theSides[i].fd(i);
	Size sum = 0;
	for (int i = 0; i < theBufs.capacity(); ++i) {
		IOBuffer *buf = new IOBuffer();
		buf->minCapacity(bufSize);
		theBufs.push(buf);
		sum += buf->bufSize();
	}
	cerr << "allocated " << sum/1024. << " KB of buffers for about " << theBufs.capacity()-1 << " connections" << endl;
}

void Select::addPair(int fd1, int fd2) {
	addFD(fd1);
	addFD(fd2);
	//cerr << " pair: fd: " << fd1 << " and " << fd2 << endl;
	Assert(theBufs.count() >= 2);
	IOBuffer *b1 = theBufs.pop(); b1->reset();
	IOBuffer *b2 = theBufs.pop(); b2->reset();
	theSides[fd1].bufs(b1, b2, "\n\r\n");
	theSides[fd2].bufs(b2, b1);
	theSides[fd1].otherSide(theSides+fd2);
	theSides[fd2].otherSide(theSides+fd1);
}

void Select::addFD(int fd) {
	Assert(fd >= 0 && fd <= theFDLimit);
	Assert(!theSides[fd].open());
	theSet[ioRead].setFD(fd);
	theSet[ioWrite].setFD(fd);
	if (fd > theMaxFD)
		theMaxFD = fd;
	theCount++;
	//cerr << "fd: " << fd << " opened; max: " << theMaxFD << "; count: " << theCount << endl;
}

void Select::delFD(int fd) {
	Assert(fd >= 0 && fd <= theFDLimit);
	Assert(theSides[fd].open());
	theSet[ioRead].clearFD(fd);
	theSet[ioWrite].clearFD(fd);
	::close(fd);
	if (fd == theMaxFD)
		while (--theMaxFD >= 0 && !theSides[theMaxFD].open());
	if (theSides[fd].alone()) {
		theBufs.push(theSides[fd].readBuf());
		theBufs.push(theSides[fd].writeBuf());
	}
	theCount--;
		
	//cerr << "fd: " << fd << " closed; max: " << theMaxFD << "; count: " << theCount << endl;
}

void Select::step() {
	if (theCount <= 0)
		return;

	//cerr << "select(" << theMaxFD << "/" << theCount << ")" << endl;

	const int count = ::select(theMaxFD+1, theSet[ioRead].readySet(), theSet[ioWrite].readySet(), 0, 0);
	if (count < 0) {
		Check(errno == EINTR);
		return;
	}

	// reads
	for (int fd = theSet[ioRead].maxFD(); fd >= 0; --fd) {
		if (theSet[ioRead].isReadyFD(fd)) {
			theSides[fd].read();
		}
	}

	// writes
	for (int fd = theSet[ioWrite].maxFD(); fd >= 0; --fd) {
		if (theSet[ioWrite].isReadyFD(fd)) {
			theSides[fd].write();
		}
	}
}

static
void Wait(int ls) {
	static fd_set s;
	FD_SET(ls, &s);
	//cerr << "waiting for connections on port " << ThePort << endl;
	select(ls+1, &s, 0, 0, 0);
}

static
void Accept(Socket ls) {
	if (!TheSelect.count())
		Wait(ls);
	do {
		Socket as = ls.accept(0, 0);
		if (as < 0)
			break;
		Socket cs;
		cs.create();
		Check(cs >= 0);
		Assert(cs.connect(TheDestHost, TheDestPort));
		TheSelect.addPair(as, cs);
	} while (1);
}

int main(int argc, char *argv[]) {
	char *h;
	if (argc != 3 || !(h = strchr(argv[2], ':'))) {
		cerr << "usage: " << argv[0] << " <listen_port> <dest_host:dest_port>" << endl;
		return -1;
	}

	ThePort = atoi(argv[1]);
	*h = '\0';
	TheDestHost = argv[2];
	TheDestPort = atoi(h+1);

	cerr << "connecting localhost:" << ThePort << " and " << TheDestHost << ':' << TheDestPort << endl;

	Socket ls;
	Check(ls.create());
	ls.blocking(false);
	ls.linger(0);
	ls.reusePort(true);
	ls.reuseAddr(true);
	Check(ls.listen(ThePort));

	SetMaxFD(GetMaxFD());
	TheSelect.limits(GetMaxFD()+1, Size::KB(4)-1);

	do {
		Accept(ls);
		TheSelect.step();
	} while (1);

	ls.close();
}
